// x509.h - Contains wrappers around OpenSSL's X509 certificate verification
#pragma once

#if defined(_WIN32)
#define API_HELPER_EXPORT __declspec(dllexport)
#else
#define API_HELPER_EXPORT __attribute__((visibility("default")))
#endif

#include <string>

namespace schrodinger::licensing
{

/// @brief VerifyOptions contains additional options for certificate
/// verification.
struct VerifyOptions {
    /// Expected EKU on the leaf certificate.
    /// For appropriate values, see the XKU_* constants in
    /// https://www.openssl.org/docs/man1.1.1/man3/X509_get_extended_key_usage.html
    ///
    /// @note A nicer API would probably re-define a list of EKUs in this
    /// header file, such as an enum, so that callers don't have to include
    /// OpenSSL headers. But, we only expect to use this within mmlic3 for now,
    /// and only XKU_CODE_SIGN.
    uint32_t ext_key_usage = 0;
};

/// @brief VerifyResult contains the result of a certificate verification.
// TODO: Consider using some result type. I like absl::Status if I had to pick.
// But, each has its own quirks, and none are conventional in Schrodinger c++
// code, yet.
struct VerifyResult {
    bool valid = false; // Did verification succeed?
    std::string error =
        "<uninitialized VerifyResult>"; // Error message if verification failed.
};

/**
 * @brief Verify the leaf certificate of a certificate chain.
 *
 * @param leaf_cert The leaf certificate to verify, as a PEM-encoded string.
 * @param cert_chain Untrusted chain of certificates that should lead to one of
 * the roots, as a PEM-encoded string.
 * @param roots A pool of trusted root certificates, as a PEM-encoded string.
 * @param options Additional options for verification, beyond chain validation.
 *
 * @return VerifyResult containing a "valid" bool, and an error message if the
 * certificate chain is invalid or leaf certificate validation fails.
 *
 * @note This function does not do revocation checking.
 *
 * @note Unlike go's crypto/x509, this does not verify the chain's EKU.
 *
 */
API_HELPER_EXPORT VerifyResult verify_certificate_chain(
    const std::string& leaf_cert, const std::string& cert_chain,
    const std::string& roots, const VerifyOptions& options);

/// @brief Supported signature algorithms for verification.
/// Naming is taken from go's crypto/x509 package:
/// https://pkg.go.dev/crypto/x509#SignatureAlgorithm
enum class SignatureAlgorithm {
    Unknown, // Invalid for use by this library
    SHA256WithRSA,
    SHA384WithRSA,
    SHA512WithRSA,
    ECDSAWithSHA256,
    ECDSAWithSHA384,
    ECDSAWithSHA512,
    SHA256WithRSAPSS,
    SHA384WithRSAPSS,
    SHA512WithRSAPSS,
    PureEd25519,
};

/**
 * @brief Convert a string to a SignatureAlgorithm. The value of the string
 * should be equal to the name of the enum.
 *
 * @returns The SignatureAlgorithm corresponding to the string, or
 * SignatureAlgorithm::Unknown if the string does not match any known algorithm.
 */
API_HELPER_EXPORT SignatureAlgorithm SignatureAlgorithmFromString(std::string_view algo);

/**
 * @brief Verify a message's cryptographic signature.
 *
 * @param signer_cert The certificate of the signer, as a PEM-encoded string.
 * @param data The exact data that was signed.
 * @param signature The exact signature of the data.
 * @param algorithm The algorithm used to sign the data.
 * @return VerifyResult
 *
 * @note `data` and `signature` are strings, but just as a container for an
 * arbitrary sequence of bytes. std::string was chosen over std::vector<uint8_t>
 * as the more likely container used by callers.
 *
 * @note This function does NOT do any validaty checking on the signer
 * certificate. Use verify_certificate_chain() for that.
 */
API_HELPER_EXPORT VerifyResult verify_signature(const std::string& signer_cert,
                                               const std::string& data,
                                               const std::string& signature,
                                               SignatureAlgorithm algorithm);

} // namespace schrodinger::licensing
